/*
AVR Matrix
 
Author: David Voswinkel
Email:  david@optixx.org
*/

#define  __AVR_AT90S8515__ 
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h> 
#include <string.h> 

#define SCALE0         251         
#define SCALE1         1      
#define BAUDRATE       (4000000/(115200*16)-1)
#define IDLE_TIMEOUT   15
#define SCROLL_STEP    6
#define FONT_WIDTH     6

 
#define ABS(A) ((A)<0?-(A):A)
#define SGN(A) (A < 0 ? -1 : (A > 0 ? 1 : 0))  


const uint8_t font[47][5] PROGMEM  = {

{0xfc,0x12,0x12,0x12,0xfc},
{0xfe,0x92,0x92,0x92,0x6c},
{0x7c,0x82,0x82,0x82,0x44},
{0xfe,0x82,0x82,0x44,0x38},
{0xfe,0x92,0x92,0x92,0x82},
{0xfe,0x12,0x12,0x12,0x02},
{0x7c,0x82,0x92,0x92,0x74},
{0xfe,0x10,0x10,0x10,0xfe},
{0x00,0x82,0xfe,0x82,0x00},
{0x60,0x80,0x80,0x80,0x7e},
{0xfe,0x10,0x28,0x44,0x82},
{0xfe,0x80,0x80,0x80,0x80},
{0xfe,0x04,0x08,0x04,0xfe},
{0xfe,0x08,0x10,0x20,0xfe},
{0x7c,0x82,0x82,0x82,0x7c},
{0xfe,0x12,0x12,0x12,0x0c},
{0x7c,0x82,0xa2,0x42,0xbc},
{0xfe,0x12,0x32,0x52,0x8c},
{0x8c,0x92,0x92,0x92,0x62},
{0x02,0x02,0xfe,0x02,0x02},
{0x7e,0x80,0x80,0x80,0x7e},
{0x3e,0x40,0x80,0x40,0x3e},
{0xfe,0x40,0x20,0x40,0xfe},
{0xc6,0x28,0x10,0x28,0xc6},
{0x06,0x08,0xf0,0x08,0x06},
{0xc2,0xa2,0x92,0x8a,0x86},
{0x7c,0xa2,0x92,0x8a,0x7c},
{0x00,0x08,0x04,0xfe,0x00},
{0xe2,0x92,0x92,0x92,0x8c},
{0x42,0x92,0x9a,0x96,0x62},
{0x10,0x18,0x14,0xfe,0x10},
{0x8e,0x92,0x92,0x92,0x62},
{0x7c,0x92,0x92,0x92,0x60},
{0x02,0xe2,0x12,0x0a,0x06},
{0x6c,0x92,0x92,0x92,0x6c},
{0x0c,0x92,0x92,0x92,0x7c},
{0xfe,0x82,0x82,0x82,0xfe},
{0x00,0x00,0x00,0x00,0xfe},
{0xf2,0x92,0x92,0x92,0x9e},
{0x92,0x92,0x92,0x92,0xfe},
{0x1e,0x10,0x10,0x10,0xfe},
{0x9e,0x92,0x92,0x92,0xf2},
{0xfe,0x92,0x92,0x92,0xf2},
{0x02,0x02,0x02,0x02,0xfe},
{0xfe,0x92,0x92,0x92,0xfe},
{0x9e,0x92,0x92,0x92,0xfe},
{0x00,0x00,0x00,0x00,0x00},
};


typedef struct framebuffer_t {
    volatile uint8_t front[8];
    volatile uint8_t data[8];
    volatile uint8_t back[8];
    volatile uint8_t col;
    volatile uint8_t vsync; 
    volatile uint8_t timer_step;
    volatile uint8_t save;
    
} framebuffer_t;

static volatile framebuffer_t __fb;

typedef struct sio_t {
    uint8_t in[30];
    uint8_t len; 
    uint8_t ready;
} sio_t;

static volatile sio_t  sio;

INTERRUPT(SIG_OVERFLOW0)
{
    TCNT0 = SCALE0;
    __fb.col++;
    
    if (__fb.col==8){
        __fb.col=0;
    }
    
    outp(__fb.data[__fb.col], PORTC);
    outp(~(1<<__fb.col), PORTB);  
}

INTERRUPT(SIG_OVERFLOW1)
{
    __fb.timer_step++;
}

SIGNAL(SIG_UART_RECV)
{
    uint8_t c;
    c  = UDR;
    if (!sio.ready){
        if (c!=0x0a) {
            sio.in[sio.len++] = c;
        } else {
            sio.ready = 1;            
        }
    }
}


void fb_init(void)
{
    outp(0xff,DDRB);            
    outp(0xff,DDRC);
    __fb.col = 0;
    __fb.vsync = 0;
    __fb.timer_step = 0,
    __fb.save = 0,
    memset((void*)__fb.front,0,64*3);
    cli();
    TIMSK =  (1<<TOIE0) | (1<<TOIE1);
    TCNT0 = SCALE0;
    TCCR0 = 5;   
    TCCR1B = 1 ; //(1<<CS11); 
    sei();
}  

inline void fb_set(uint8_t x,uint8_t y)
{
    __fb.data[x&7] |= (1<<(y&7)); 
}

void fb_line(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2)
{
    
    int8_t d, x, y, ax, ay, sx, sy, dx, dy;

    dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
    dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

    x = x1;
    y = y1;
    if (ax>ay) {		
        d = ay-(ax>>1);
        for (;;) {
            fb_set(x, y);
            if (x==x2) return;
            if (d>=0) {
                y += sy;
                d -= ax;
            }
            x += sx;
            d += ay;
        }
    }
    else {	
        d = ax-(ay>>1);
        for (;;) {
            fb_set(x, y);
            if (y==y2) return;
            if (d>=0) {
                x += sx;
                d -= ay;
            }
            y += sy;
            d += ax;
        }
    }
}

inline void fb_rect(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2)
{
    fb_line(x1,y1,x2,y1);
    fb_line(x2,y1,x2,y2);
    fb_line(x2,y2,x1,y2);
    fb_line(x1,y2,x1,y1);
}

void fb_circle(uint8_t x, uint8_t  y, uint8_t  radius)
{
    int8_t  balance, xoff, yoff;
    
    xoff = 0;
    yoff = radius;
    balance = -radius;
    
    do {
        fb_set(x+xoff, y+yoff);
        fb_set(x-xoff, y+yoff);
        fb_set(x-xoff, y-yoff);
        fb_set(x+xoff, y-yoff);
        fb_set(x+yoff, y+xoff);
        fb_set(x-yoff, y+xoff);
        fb_set(x-yoff, y-xoff);
        fb_set(x+yoff, y-xoff);
        balance += xoff++;
        if (balance + xoff  >= 0){
            --yoff;
            balance -= yoff + yoff;
        }
    } while (xoff <= yoff);
} 

inline void fb_clear(void)
{
    memset((void*)__fb.front,0,64*3);
}

inline uint8_t fb_ticked(void)
{
    if (__fb.timer_step != __fb.save) {
        __fb.save = __fb.timer_step;
        return 1;
    }
    return 0;
}

void fb_putc(uint8_t c,uint8_t *  fb)
{
    if (c!=32)
        c = c - 65;
    else 
        c = 46;
    memcpy_P((void*)fb,font + c ,5); 
}

void fb_scroll_left(void)
{
  memmove((void*)__fb.data,(const void*)__fb.data+1,15);
}

void fb_scroll_right(void)
{
    memmove((void*)__fb.front+1,(const void*)__fb.front,15);
}


void fb_roll_left(void)
{
    uint8_t save = __fb.data[0];
    memmove((void*)__fb.data,(const void*)__fb.data+1,7);
    __fb.data[7] = save;          

}

void fb_roll_right(void)
{
    uint8_t save = __fb.data[7];
    memmove((void*)__fb.data+1,(const void*)__fb.data,7);
    __fb.data[0] = save;          
}

void fb_shift_left(void)
{
    memmove((void*)__fb.data,(const void*)__fb.data+1,7);
}

void fb_shift_right(void)
{
    memmove((void*)__fb.data+1,(const void*)__fb.data,7);
}

void fb_shift_up(void)
{
    uint8_t col;
    for (col=0; col<8;col++)
        __fb.data[col] = __fb.data[col] >> 1;       
}

void fb_shift_down(void)
{
    uint8_t col;
    for (col=0; col<8;col++)
        __fb.data[col] = __fb.data[col] << 1;       
}

void fb_roll_up(void)
{
    uint8_t col;
    for (col=0; col<8;col++){
        __fb.data[col] = __fb.data[col] >> 1 | __fb.data[col] << 7;       
       }
}

void fb_roll_down(void)
{
    uint8_t col;
    for (col=0; col<8;col++)
        __fb.data[col] = __fb.data[col] << 1 | __fb.data[col] >> 7;       
}


void sio_init(void)
{
    
    memset((void*)sio.in,0,30);
    sio.len = 0;
    sio.ready = 0;
    cli();    
    outp(0xff,DDRD);          
    UCR = BV(RXEN)|BV(RXCIE)|BV(RXB8); 
    UBRR  = BAUDRATE;
    sei();    
    
}

uint8_t text[] PROGMEM = "AVR MATRIX XMMS PLUGIN   ";

// echo -en "\x01\x03\x07\x0f\x1f\x3f\x7f\xff\x0a" > /dev/ttyS0


int main( void )
{
    uint8_t c,t,x = 0;
    uint8_t idle = 0;
    uint16_t index = 0;
    sio_init();
    fb_init();
    fb_clear();
    for (;;) {
        if ( fb_ticked() ) {
            if (sio.ready) {
                memcpy((void*)__fb.data,(void*)sio.in,8);
                sio.len   = 0;
                sio.ready = 0;
                idle = 0;
            } else {
                if (idle < IDLE_TIMEOUT) {
                    idle++;
                } else if (IDLE_TIMEOUT == idle) {
                    index=x=t=0;
                    fb_clear();                
                } else {
                    t++;
                    if (t >= SCROLL_STEP){
                        t = 0;
                        fb_scroll_left();
                        x++;
                        if (x >= FONT_WIDTH) {
                            x=0;
                            c =  pgm_read_byte(text + index++);
                            if (!c)
                                index = 0;
                            else    
                                fb_putc(c,(uint8_t*)__fb.back);
                        }
                    }
                }
            }
        }
    }
}


